# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

import os.path
import sys
from unittest import mock

import pytest

from antsibull_core import subprocess_util
from antsibull_core.venv import FakeVenvRunner, VenvRunner, get_clean_environment


def test_venv_clean_env(monkeypatch):
    monkeypatch.setenv('PYTHONPATH', '/jfjfjfjfjfjfjfjfj')
    assert 'PYTHONPATH' not in get_clean_environment()


def test_venv_run_init(tmp_path):
    with mock.patch('antsibull_core.subprocess_util.async_log_run') as log_run:
        runner = VenvRunner('asdfgh', tmp_path)
        assert runner.name == 'asdfgh'
        assert runner.top_dir == tmp_path
        assert runner.venv_dir == str(tmp_path / 'asdfgh')
        pip = os.path.join(runner.venv_dir, 'bin', 'pip')
        log_run.assert_called_once_with(
            [pip, 'install', '--upgrade', 'pip'],
            None,
            None,
            'debug',
            True,
            errors='strict',
            env=get_clean_environment(),
        )


def test_venv_log_run_error(tmp_path):
    runner = VenvRunner('zxcvb', tmp_path)
    echo = os.path.join(runner.venv_dir, 'bin', 'echo')
    with pytest.raises(ValueError, match=rf'^{echo!r} does not exist!'):
        runner.log_run(['echo', "This won't work!"])


def test_venv_log_run_error2(tmp_path):
    runner = VenvRunner('zxcvb', tmp_path)
    echo = '/usr/bin/echo'
    with pytest.raises(ValueError, match=rf'^{echo!r} must not be an absolute path!'):
        runner.log_run([echo, "This also won't work!"])


def test_fake_venv_python():
    with mock.patch.object(
        subprocess_util, 'async_log_run', wraps=subprocess_util.async_log_run
    ) as log_run:
        args = ['python', '-c', 'import antsibull_core; print("Hello, world!")']
        p = FakeVenvRunner().log_run(args)
        assert p.stdout == 'Hello, world!\n'

        args = args.copy()
        args[0] = sys.executable
        log_run.assert_called_once_with(
            args,
            None,
            None,
            'debug',
            True,
            errors='strict',
        )



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

import hashlib

import aiohttp
import pytest

from antsibull_core.galaxy import CollectionDownloader, GalaxyClient, GalaxyContext, GalaxyVersion

KNOWN_CG_VERSIONS = [
    '6.4.0', '6.3.0', '6.2.0', '6.1.0', '6.0.1', '6.0.0', '6.0.0-a1',
    '5.8.6', '5.8.5', '5.8.4', '5.8.3', '5.8.2', '5.8.1', '5.8.0', '5.7.0', '5.6.0', '5.5.0', '5.4.0', '5.3.0', '5.2.0', '5.1.1', '5.1.0', '5.0.2', '5.0.1', '5.0.0', '5.0.0-a1',
    '4.8.9', '4.8.8', '4.8.7', '4.8.6', '4.8.5', '4.8.4', '4.8.3', '4.8.2', '4.8.1', '4.8.0', '4.7.0', '4.6.1', '4.6.0', '4.5.0', '4.4.0', '4.3.0', '4.2.0', '4.1.0', '4.0.2', '4.0.1', '4.0.0',
    '3.8.10', '3.8.9', '3.8.8', '3.8.7', '3.8.6', '3.8.5', '3.8.4', '3.8.3', '3.8.2', '3.8.1', '3.8.0', '3.7.0', '3.6.0', '3.5.0', '3.4.0', '3.3.2', '3.3.1', '3.3.0', '3.2.0', '3.1.0', '3.0.2', '3.0.1', '3.0.0',
    '2.5.9', '2.5.8', '2.5.7', '2.5.6', '2.5.5', '2.5.4', '2.5.3', '2.5.2', '2.5.1', '2.5.0', '2.4.0', '2.3.0', '2.2.0', '2.1.1', '2.1.0', '2.0.1', '2.0.0',
    '1.3.14', '1.3.13', '1.3.12', '1.3.11', '1.3.10', '1.3.9', '1.3.8', '1.3.7', '1.3.6', '1.3.5', '1.3.4', '1.3.3', '1.3.2', '1.3.1', '1.3.0', '1.2.0', '1.1.0', '1.0.0',
    '0.3.0-experimental.meta.redirects-3', '0.3.0-experimental.meta.redirects-2', '0.3.0-experimental.meta.redirects', '0.2.1', '0.2.0', '0.1.4', '0.1.1',
]


async def galaxy_client_test(aio_session: aiohttp.ClientSession, context: GalaxyContext,
                             tmp_path_factory,
                             skip_versions_test: bool = False,
                             skip_download_test: bool = False,
                             ) -> None:
    client = GalaxyClient(aio_session, context=context)

    # Check collection info
    cg_info = await client.get_info('community.general')
    print(cg_info)
    assert cg_info['name'] == 'general'
    assert cg_info['deprecated'] is False

    # Check info on specific collection version
    cg_version = await client.get_release_info('community.general', '6.0.0')
    cg_version.pop('files', None)  # this is huge for Galaxy v3 API
    cg_version.pop('metadata', None)  # this is large for Galaxy v3 API
    print(cg_version)
    assert 'download_url' in cg_version
    assert cg_version['artifact']['filename'] == 'community-general-6.0.0.tar.gz'
    assert cg_version['artifact']['size'] == 2285782
    expected = '7c7ec856355078577b520f7432645754d75ad8e74a46e84d1ffee8fad80efc5c'
    assert cg_version['artifact']['sha256'] == expected
    assert cg_version['namespace']['name'] == 'community'
    assert cg_version['collection']['name'] == 'general'
    assert cg_version['version'] == '6.0.0'

    # Check collection list
    if not skip_versions_test:
        cg_versions = await client.get_versions('community.general')
        print(cg_versions)
        for known_version in KNOWN_CG_VERSIONS:
            assert known_version in cg_versions

    # Download collection
    if not skip_download_test:
        download_path = tmp_path_factory.mktemp('download')
        cache_path = tmp_path_factory.mktemp('cache')
        downloader = CollectionDownloader(
            aio_session,
            str(download_path),
            collection_cache=str(cache_path),
            context=context,
        )
        path = await downloader.download('community.dns', '0.1.0')
        with open(path, 'rb') as f:
            data = f.read()
        length = len(data)
        print(length)
        assert length == 133242
        m = hashlib.sha256()
        m.update(data)
        digest = m.hexdigest()
        expected = '2de9d40940536e65b35995a3f58dea7776de3f23f1a7ab865e0b3b8482d746b5'
        assert digest == expected

        # Try again, should hit cache
        path2 = await downloader.download('community.dns', '0.1.0')
        with open(path2, 'rb') as f:
            data2 = f.read()
        assert path == path2
        assert data == data2


@pytest.mark.asyncio
async def test_galaxy_v2(tmp_path_factory):
    galaxy_url = 'https://galaxy.ansible.com'
    async with aiohttp.ClientSession() as aio_session:
        context = await GalaxyContext.create(aio_session, galaxy_url)
        assert context.version == GalaxyVersion.V2
        assert context.base_url == galaxy_url + '/api/v2/'
        await galaxy_client_test(aio_session, context, tmp_path_factory)


@pytest.mark.asyncio
async def test_galaxy_v3(tmp_path_factory):
    galaxy_url = 'https://beta-galaxy.ansible.com'
    async with aiohttp.ClientSession() as aio_session:
        context = await GalaxyContext.create(aio_session, galaxy_url)
        assert context.version == GalaxyVersion.V3
        assert context.base_url == galaxy_url + '/api/v3/'
        await galaxy_client_test(aio_session, context, tmp_path_factory, skip_versions_test=True)



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

from unittest.mock import MagicMock, call

import pytest

import antsibull_core.subprocess_util


def test_log_run() -> None:
    logger = MagicMock()
    args = ('bash', '-ec', 'echo 123 && echo 456 >&2')
    proc = antsibull_core.subprocess_util.log_run(args, logger)
    assert proc.args == args
    assert proc.returncode == 0
    assert proc.stdout == '123\n'
    assert proc.stderr == '456\n'
    calls = [call(f'Running subprocess: {args}'), call('stderr: 456')]
    assert logger.debug.call_args_list == calls


def test_log_run_multi() -> None:
    logger = MagicMock()
    command = """
    for i in {1..15}; do
        echo "$i: Hello, stdout"
        echo "$i: Hello, stderr" >&2
    done
    """
    args = ('bash', '-ec', command)
    proc = antsibull_core.subprocess_util.log_run(args, logger, 'info', 'warn')
    assert proc.args == args
    assert proc.returncode == 0

    assert logger.debug.called_once_with(f'Running subprocess: {args}')
    assert logger.warn.call_count == 15
    assert logger.info.call_count == 15
    expected_out: list[str] = []
    expected_err: list[str] = []
    for index, inp in enumerate(logger.info.call_args_list):
        msg = f'{index+1}: Hello, stdout'
        expected_out.append(msg)
        assert inp == call('stdout: ' + msg)
    for index, inp in enumerate(logger.warn.call_args_list):
        msg = f'{index+1}: Hello, stderr'
        expected_err.append(msg)
        assert inp == call('stderr: ' + msg)
    assert proc.stdout == '\n'.join(expected_out) + '\n'
    assert proc.stderr == '\n'.join(expected_err) + '\n'


@pytest.mark.parametrize('count', [
    8 * 1024 * 1024 - 1,  # should not trigger long line code
    8 * 1024 * 1024,  # should not trigger long line code
    8 * 1024 * 1024 + 1,
    8 * 1024 * 1024 + 10,
    9 * 1024 * 1024,
])
def test_log_run_long_line(count: int) -> None:
    args = ('sh', '-c', f'dd if=/dev/zero of=/dev/stdout bs={count} count=1 ; echo ; echo foo')
    proc = antsibull_core.subprocess_util.log_run(args)
    assert proc.args == args
    assert proc.returncode == 0
    assert proc.stdout == ('\u0000' * count) + '\nfoo\n'



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Utilities for dealing with subprocesses
"""

from __future__ import annotations

import asyncio
import subprocess
from asyncio.exceptions import IncompleteReadError, LimitOverrunError
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, cast

from antsibull_core.logging import log

if TYPE_CHECKING:
    from logging import Logger as StdLogger

    from _typeshed import StrOrBytesPath
    from twiggy.logger import Logger as TwiggyLogger  # type: ignore[import]

mlog = log.fields(mod=__name__)

CalledProcessError = subprocess.CalledProcessError


async def _stream_log(
    name: str, callback: Callable[[str], Any] | None, stream: asyncio.StreamReader,
    errors: str,
) -> str:
    # We do not simply use stream.readline() since it has a line length limit.
    # While we set this limit already to 8 MB (the default is 64 KB), we still
    # want to cover longer lines as well, so we use stream.readuntil('\n')
    # and manually handle the case of longer lines.
    lines = []
    line_parts = []
    sep = b'\n'
    while True:
        try:
            line_parts.append(await stream.readuntil(sep))
        except IncompleteReadError as e:
            line_parts.append(e.partial)
        except LimitOverrunError as e:
            part = await stream.read(e.consumed)
            line_parts.append(part)
            if part:
                continue

        line = b''.join(line_parts)
        line_parts.clear()
        if not line:
            break
        text = line.decode('utf-8', errors=errors)
        if callback:
            callback(f'{name}: {text.strip()}')
        lines.append(text)
    return ''.join(lines)


async def async_log_run(
    args: Sequence[StrOrBytesPath],
    logger: TwiggyLogger | StdLogger | None = None,
    stdout_loglevel: str | None = None,
    stderr_loglevel: str | None = 'debug',
    check: bool = True,
    *,
    errors: str = 'strict',
    **kwargs,
) -> subprocess.CompletedProcess[str]:
    """
    Asynchronously run a command in a subprocess and log its output.
    The command's stdout and stderr are always captured.
    For some usecases, you may still need to call
    asyncio.create_subprocess_exec() directly to have more control.

    :param args: Command to run
    :param logger:
        Logger in which to log the command. Can be a `twiggy.logger.Logger` or
        a stdlib `logger.Logger`.
    :param stdout_loglevel: Which level to use to log stdout. `None` disables logging.
    :param stderr_loglevel: Which level to use to log stderr. `None` disables logging.
    :param check:
        Whether to raise a `subprocess.CalledProcessError` when the
        command returns a non-zero exit code
    :param errors:
        How to handle UTF-8 decoding errors. Default is ``strict``.
    """
    logger = logger or mlog
    stdout_logfunc: Callable[[str], Any] | None = None
    if stdout_loglevel:
        stdout_logfunc = getattr(logger, stdout_loglevel)
    stderr_logfunc: Callable[[str], Any] | None = None
    if stderr_loglevel:
        stderr_logfunc = getattr(logger, stderr_loglevel)

    logger.debug(f'Running subprocess: {args!r}')
    kwargs['stdout'] = asyncio.subprocess.PIPE
    kwargs['stderr'] = asyncio.subprocess.PIPE
    kwargs['limit'] = 2 ** 23  # Increase line length limit to 8 MB (the default is 64k)
    proc = await asyncio.create_subprocess_exec(*args, **kwargs)
    stdout, stderr = await asyncio.gather(
        # proc.stdout and proc.stderr won't be None with PIPE, hence the cast()
        asyncio.create_task(
            _stream_log(
                'stdout',
                stdout_logfunc,
                cast(asyncio.StreamReader, proc.stdout),
                errors,
            )
        ),
        asyncio.create_task(
            _stream_log(
                'stderr',
                stderr_logfunc,
                cast(asyncio.StreamReader, proc.stderr),
                errors,
            )
        ),
    )
    returncode = await proc.wait()

    completed = subprocess.CompletedProcess(
        args=args, returncode=returncode, stdout=stdout, stderr=stderr
    )
    if check:
        completed.check_returncode()
    return completed


def log_run(
    args: Sequence[StrOrBytesPath],
    logger: TwiggyLogger | StdLogger | None = None,
    stdout_loglevel: str | None = None,
    stderr_loglevel: str | None = 'debug',
    check: bool = True,
    **kwargs,
) -> subprocess.CompletedProcess[str]:
    """
    Synchronous wrapper for the async_log_run function.
    This function runs a command in a subprocess and captures and logs its
    output.
    """
    return asyncio.run(
        async_log_run(args, logger, stdout_loglevel, stderr_loglevel, check, **kwargs)
    )


__all__ = ('async_log_run', 'log_run', 'CalledProcessError')



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Functions to work with Galaxy."""

from __future__ import annotations

import asyncio
import dataclasses
import os.path
import shutil
import typing as t
from enum import Enum
from urllib.parse import urljoin

import aiofiles
import semantic_version as semver

from . import app_context
from .utils.hashing import verify_hash
from .utils.http import retry_get

# The type checker can handle finding aiohttp.client but flake8 cannot :-(
if t.TYPE_CHECKING:
    import aiohttp.client


#: URL to galaxy. (Get the default from the context default)
_GALAXY_SERVER_URL = str(app_context.AppContext().galaxy_url)


class NoSuchCollection(Exception):
    """Collection name does not map to a collection on Galaxy."""


class NoSuchVersion(Exception):
    """Version does not match with any versions of a collection on Galaxy."""


class DownloadFailure(Exception):
    """Failure downloading a collection from Galaxy."""


class DownloadResults(t.NamedTuple):
    """Results of downloading a collection."""

    #: :obj:`semantic_version.Version` of the exact version of the collection that was downloaded.
    version: semver.Version
    #: Location on the filesystem of the downloaded collection.
    download_path: str


class GalaxyVersion(Enum):
    V2 = 2
    V3 = 3


@dataclasses.dataclass
class GalaxyContext:
    server: str
    version: GalaxyVersion
    base_url: str

    @classmethod
    async def create(cls, aio_session: aiohttp.client.ClientSession, galaxy_server: str
                     ) -> GalaxyContext:
        api_url = urljoin(galaxy_server, 'api/')
        async with retry_get(aio_session, api_url,
                             headers={'Accept': 'application/json'}) as response:
            galaxy_info = await response.json()
        available_versions: t.Mapping[str, str] = galaxy_info.get('available_versions') or {}
        if 'v3' in available_versions:
            version = GalaxyVersion.V3
            base_url = urljoin(galaxy_server, 'api/' + available_versions['v3'])
        elif 'v2' in available_versions:
            version = GalaxyVersion.V2
            base_url = urljoin(galaxy_server, 'api/' + available_versions['v2'])
        else:
            raise RuntimeError(
                f'Information retrieved from {api_url} seems to indicate'
                ' neither Galaxy v2 API nor Galaxy v3 API'
            )
        return cls(galaxy_server, version, base_url)


_GALAXY_CONTEXT_CACHE: dict[str, t.Union[GalaxyContext, asyncio.Future]] = {}


async def _get_cached_galaxy_context(aio_session: aiohttp.client.ClientSession,
                                     galaxy_server: str) -> GalaxyContext:
    context_or_future = _GALAXY_CONTEXT_CACHE.get(galaxy_server)
    if context_or_future is not None:
        if asyncio.isfuture(context_or_future):
            return await context_or_future
        return t.cast(GalaxyContext, context_or_future)

    loop = asyncio.get_running_loop()
    future = loop.create_future()

    async def _init():
        try:
            context = await GalaxyContext.create(aio_session, galaxy_server)
            future.set_result(context)
            _GALAXY_CONTEXT_CACHE[galaxy_server] = context
        except Exception as exc:  # pylint: disable=broad-exception-caught
            future.set_exception(exc)

    loop.create_task(_init())
    _GALAXY_CONTEXT_CACHE[galaxy_server] = future
    return await future


class GalaxyClient:
    """Class for querying the Galaxy REST API."""

    def __init__(self, aio_session: aiohttp.client.ClientSession,
                 galaxy_server: t.Optional[str] = None,
                 context: t.Optional[GalaxyContext] = None) -> None:
        """
        Create a GalaxyClient object to query the Galaxy Server.

        :arg aio_session: :obj:`aiohttp.ClientSession` with which to perform all
            requests to galaxy.
        :kwarg galaxy_server: URL to the galaxy server. ``context`` must be provided instead
            in the future.
        :kwarg context: A ``GalaxyContext`` instance. Must be provided in the future.
        """
        if galaxy_server is None and context is None:
            # TODO: deprecate
            galaxy_server = _GALAXY_SERVER_URL
        elif context is not None:
            # TODO: deprecate
            if galaxy_server is not None and galaxy_server != context.server:
                raise ValueError(
                    f'galaxy_server ({galaxy_server}) does not coincide'
                    f' with context.server ({context.server})'
                )
            galaxy_server = context.server
        self.galaxy_server = galaxy_server
        self.context = context
        self.aio_session = aio_session
        self.headers: dict[str, str] = {'Accept': 'application/json'}
        self.params: dict[str, str] = {}
        if context:
            self._update_from_context(context)

    def _update_from_context(self, context: GalaxyContext) -> None:
        if context.version == GalaxyVersion.V2:
            self.params['format'] = 'json'

    async def _ensure_context(self) -> GalaxyContext:
        """
        Ensure that ``self.context`` is present.
        """
        context = self.context
        if context is not None:
            return context
        if self.galaxy_server is None:
            raise RuntimeError('Unexpected None for GalaxyClient.galaxy_server')
        context = await _get_cached_galaxy_context(self.aio_session, self.galaxy_server)
        self.context = context
        self._update_from_context(context)
        return context

    async def _get_galaxy_versions(self, context: GalaxyContext, versions_url: str,
                                   add_params: bool = True) -> list[str]:
        """
        Retrieve the complete list of versions for a collection from a galaxy endpoint.

        This internal function retrieves versions for collections from a Galaxy endpoint.  If the
        information is paged, it continues to retrieve linked pages until all of the information has
        been returned.

        :arg context: the ``GalaxyContext`` to use.
        :arg version_url: url to the page to retrieve.
        :arg add_params: used internally during recursion. Do not specify when calling this.
        :returns: List of the all the versions of the collection.
        """
        if add_params:
            params = self.params.copy()
            if context.version == GalaxyVersion.V2:
                params['page_size'] = '100'
            else:
                params['limit'] = '50'
        else:
            params = None
        async with retry_get(self.aio_session, versions_url, params=params,
                             headers=self.headers, acceptable_error_codes=[404]) as response:
            if response.status == 404:
                raise NoSuchCollection(f'No collection found at: {versions_url}')
            collection_info = await response.json()

        versions = []
        if context.version == GalaxyVersion.V2:
            results = collection_info['results']
            next_link = collection_info['next']
        else:
            if 'data' in collection_info:
                # Apparently 'data' isn't always used...
                results = collection_info['data']
            else:
                results = collection_info['results']
            next_link = collection_info['links']['next']
            add_params = False
        for version_record in results:
            versions.append(version_record['version'])
        if next_link:
            if next_link.startswith('/'):
                next_link = urljoin(context.server, next_link)
            versions.extend(await self._get_galaxy_versions(context, next_link, add_params))

        return versions

    async def get_versions(self, collection: str) -> list[str]:
        """
        Retrieve all versions of a collection on Galaxy.

        :arg collection: Name of the collection to get version info for.
        :returns: List of all the versions of this collection on galaxy.
        """
        context = await self._ensure_context()

        collection = collection.replace('.', '/')
        galaxy_url = urljoin(context.base_url, f'collections/{collection}/versions/')
        retval = await self._get_galaxy_versions(context, galaxy_url)
        return retval

    async def get_info(self, collection: str) -> dict[str, t.Any]:
        """
        Retrieve information about the collection on Galaxy.

        :arg collection: Namespace.collection to retrieve information about.
        :returns: Dictionary of information about the collection.

        Please see the Galaxy v2 and v3 REST API documentation for information on the
        structure of the returned data.

        .. seealso::
            An example return value from the
            `Galaxy v2 REST API
            <https://galaxy.ansible.com/api/v2/collections/community/general/>`_
            and the `Galaxy v3 REST API
            <https://beta-galaxy.ansible.com/api/v3/plugin/ansible/content/published/collections/index/community/general/>`_
        """
        context = await self._ensure_context()

        collection = collection.replace('.', '/')
        galaxy_url = urljoin(context.base_url, f'collections/{collection}/')

        async with retry_get(self.aio_session, galaxy_url, params=self.params,
                             headers=self.headers, acceptable_error_codes=[404]) as response:
            if response.status == 404:
                raise NoSuchCollection(f'No collection found at: {galaxy_url}')
            collection_info = await response.json()

        return collection_info

    async def get_release_info(self, collection: str,
                               version: str | semver.Version) -> dict[str, t.Any]:
        """
        Retrive information about a specific version of a collection.

        :arg collection: Namespace.collection string naming the collection.
        :arg version: Version of the collection.
        :returns: Dictionary of information about the release.

        Please see the Galaxy v2 and v3 REST API documentation for information on the
        structure of the returned data.

        .. seealso::
            An example return value from the
            `Galaxy v2 REST API
            <https://galaxy.ansible.com/api/v2/collections/community/general/versions/0.1.1>`_
            and the `Galaxy v3 REST API
            <https://beta-galaxy.ansible.com/api/v3/plugin/ansible/content/published/collections/index/community/general/versions/0.1.1/>`_
        """
        context = await self._ensure_context()

        collection = collection.replace('.', '/')
        galaxy_url = urljoin(context.base_url, f'collections/{collection}/versions/{version}/')

        async with retry_get(self.aio_session, galaxy_url, params=self.params,
                             headers=self.headers, acceptable_error_codes=[404]) as response:
            if response.status == 404:
                raise NoSuchCollection(f'No collection found at: {galaxy_url}')
            collection_info = await response.json()

        return collection_info

    async def get_latest_matching_version(self, collection: str,
                                          version_spec: str,
                                          pre: bool = False) -> semver.Version:
        """
        Get the latest version of a collection that matches a specification.

        :arg collection: Namespace.collection identifying a collection.
        :arg version_spec: String specifying the allowable versions.
        :kwarg pre: If True, allow prereleases (versions which have the form X.Y.Z.SOMETHING).
            This is **not** for excluding 0.Y.Z versions.  non-pre-releases are still
            preferred over pre-releases (for instance, with version_spec='>2.0.0-a1,<3.0.0'
            and pre=True, if the available versions are 2.0.0-a1 and 2.0.0-a2, then 2.0.0-a2
            will be returned.  If the available versions are 2.0.0 and 2.1.0-b2, 2.0.0 will be
            returned since non-pre-releases are preferred.  The default is False
        :returns: :obj:`semantic_version.Version` of the latest collection version that satisfied
            the specification.

        .. seealso:: For the format of the version_spec, see the documentation
            of :obj:`semantic_version.SimpleSpec`

        .. versionchanged:: 0.37.0
            Giving True to the ``pre`` parameter now means that prereleases will be
            *allowed* but stable releases will still be *preferred*.  Previously, the
            latest release, whether stable or prerelease was returned when pre was True.
        """
        versions = await self.get_versions(collection)
        sem_versions = [semver.Version(v) for v in versions]
        sem_versions.sort(reverse=True)

        spec = semver.SimpleSpec(version_spec)
        prereleases = []
        for version in (v for v in sem_versions if v in spec):
            # If this is a pre-release, first check if there's a non-pre-release that
            # will satisfy the version_spec.
            if version.prerelease:
                prereleases.append(version)
                continue
            return version

        # We did not find a stable version that satisies the version_spec.  If we
        # allow prereleases, return the latest of those here.
        if pre and prereleases:
            return prereleases[0]

        # No matching versions were found
        raise NoSuchVersion(f'{version_spec} did not match with any version of {collection}.')


class CollectionDownloader(GalaxyClient):
    """Manage downloading collections from Galaxy."""

    def __init__(self, aio_session: aiohttp.client.ClientSession,
                 download_dir: str,
                 galaxy_server: t.Optional[str] = None,
                 collection_cache: str | None = None,
                 context: t.Optional[GalaxyContext] = None) -> None:
        """
        Create an object to download collections from galaxy.

        :arg aio_session: :obj:`aiohttp.ClientSession` with which to perform all
            requests to galaxy.
        :arg download_dir: Directory to download into.
        :kwarg galaxy_server: URL to the galaxy server. ``context`` must be provided instead
            in the future.
        :kwarg context: A ``GalaxyContext`` instance. Must be provided in the future.
        :kwarg collection_cache: If given, a path to a directory containing collection tarballs.
            These tarballs will be used instead of downloading new tarballs provided that the
            versions match the criteria (latest compatible version known to galaxy).
        """
        super().__init__(aio_session, galaxy_server=galaxy_server, context=context)
        self.download_dir = download_dir
        self.collection_cache: t.Final[str | None] = collection_cache

    async def download(self, collection: str, version: str | semver.Version, ) -> str:
        """
        Download a collection.

        Downloads the collection at the specified version.

        :arg collection: Namespace.collection identifying the collection.
        :arg version: Version of the collection to download.
        :returns: The full path to the downloaded collection.
        """
        collection = collection.replace('.', '/')
        release_info = await self.get_release_info(collection, version)
        release_url = release_info['download_url']

        download_filename = os.path.join(self.download_dir, release_info['artifact']['filename'])
        sha256sum = release_info['artifact']['sha256']

        if self.collection_cache:
            if release_info['artifact']['filename'] in os.listdir(self.collection_cache):
                cached_copy = os.path.join(self.collection_cache,
                                           release_info['artifact']['filename'])
                if await verify_hash(cached_copy, sha256sum):
                    shutil.copyfile(cached_copy, download_filename)
                return download_filename

        async with retry_get(self.aio_session, release_url,
                             acceptable_error_codes=[404]) as response:
            if response.status == 404:
                raise NoSuchCollection(f'No collection found at: {release_url}')

            async with aiofiles.open(download_filename, 'wb') as f:
                lib_ctx = app_context.lib_ctx.get()
                while chunk := await response.content.read(lib_ctx.chunksize):
                    await f.write(chunk)

        # Verify the download
        if not await verify_hash(download_filename, sha256sum):
            raise DownloadFailure(f'{release_url} failed to download correctly.'
                                  f' Expected checksum: {sha256sum}')

        # Copy downloaded collection into cache
        if self.collection_cache:
            cached_copy = os.path.join(self.collection_cache,
                                       release_info['artifact']['filename'])
            shutil.copyfile(download_filename, cached_copy)

        return download_filename

    async def download_latest_matching(self, collection: str,
                                       version_spec: str) -> DownloadResults:
        """
        Download the latest version of a collection that matches a specification.

        :arg collection: Namespace.collection identifying a collection.
        :arg version_spec: String specifying the allowable versions.
        :returns: :obj:`DownloadResults` with version and download path for the collection we
            downloaded.

        .. seealso:: For the format of the version_spec, see the documentation
            of :obj:`semantic_version.SimpleSpec`
        """
        version = await self.get_latest_matching_version(collection, version_spec)
        download_path = await self.download(collection, version)
        return DownloadResults(version=version, download_path=download_path)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Compat for older versions of Python.

PARTS OF THIS MODULE ARE DEPRECATED AND WILL BE REMOVED IN ANTSIBULL-CORE 3.0.0:
  - the `metadata` module;
  - the functions `asyncio_run`, `best_get_loop`, and `create_task`.
"""

from __future__ import annotations

import argparse
import asyncio
import sys
from importlib import metadata

BooleanOptionalAction: type[argparse.BooleanOptionalAction]

if sys.version_info < (3, 9, 11) or (
    sys.version_info >= (3, 10, 0) and sys.version_info < (3, 10, 3)
):
    # https://bugs.python.org/issue46080 was fixed in Python 3.11.0 alpha 5
    # (https://docs.python.org/3/whatsnew/changelog.html#python-3-11-0-alpha-5)
    # and backported to Python 3.10.3
    # (https://docs.python.org/3.10/whatsnew/changelog.html#python-3-10-3-final)
    # and Python 3.9.11
    # (https://docs.python.org/3.9/whatsnew/changelog.html#python-3-9-11-final).
    # Versions before these have to use the vendored version so that users can actually
    # use `--help` for affected subcommands.
    from .vendored._argparse_booleanoptionalaction import BooleanOptionalAction  # type: ignore
else:
    BooleanOptionalAction = argparse.BooleanOptionalAction


best_get_loop = asyncio.get_running_loop
asyncio_run = asyncio.run
create_task = asyncio.create_task

__all__ = ('BooleanOptionalAction', 'asyncio_run', 'best_get_loop', 'create_task', 'metadata')



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2021, Toshio Kuratomi
"""Schemas for app and lib contexts."""

import typing as t

import pydantic as p

from ..utils.collections import ContextDict
from .config import DEFAULT_LOGGING_CONFIG, LoggingModel
from .validators import convert_none, convert_path


class BaseModel(p.BaseModel):
    """
    Configuration for all Context object classes.

    :cvar Config: Sets the following information

        :cvar allow_mutation: ``False``.  Prevents setattr on the contexts.
        :cvar extra: ``p.Extra.forbid``.  Prevents extra fields on the contexts.
        :cvar validate_all: ``True``.  Validates default values as well as user supplied ones.
    """

    class Config:
        """
        Set default configuration for building the context models.

        :cvar allow_mutation: ``False``.  Prevents setattr on the contexts.
        :cvar extra: ``p.Extra.forbid``.  Prevents extra fields on the contexts.
        :cvar validate_all: ``True``.  Validates default values as well as user supplied ones.
        """

        allow_mutation = False
        extra = p.Extra.forbid
        validate_all = True


class AppContext(BaseModel):
    """
    Structure and defaults of the app_ctx.

    :ivar extra: a mapping of arg/config keys to values.  Anything in here is unchecked by a
        schema.  These are usually leftover command line arguments and config entries. If
        values stored in extras need default values, they need to be set outside of the context
        or the entries can be given an actual entry in the AppContext to take advantage of the
        schema's checking, normalization, and default setting.
    :ivar ansible_base_url: Url to the ansible-core git repo.
    :ivar galaxy_url: URL of the galaxy server to get collection info from
    :ivar logging_cfg: Configuration of the application logging
    :ivar pypi_url: URL of the pypi server to query for information
    :ivar collection_cache: If set, must be a path pointing to a directory where collection
        tarballs are cached so they do not need to be downloaded from Galaxy twice.
    """

    extra: ContextDict = ContextDict()
    # pyre-ignore[8]: https://github.com/samuelcolvin/pydantic/issues/1684
    ansible_base_url: p.HttpUrl = 'https://github.com/ansible/ansible/'  # type: ignore[assignment]
    # pyre-ignore[8]: https://github.com/samuelcolvin/pydantic/issues/1684
    galaxy_url: p.HttpUrl = 'https://galaxy.ansible.com/'  # type: ignore[assignment]
    logging_cfg: LoggingModel = LoggingModel.parse_obj(DEFAULT_LOGGING_CONFIG)
    # pyre-ignore[8]: https://github.com/samuelcolvin/pydantic/issues/1684
    pypi_url: p.HttpUrl = 'https://pypi.org/'  # type: ignore[assignment]
    collection_cache: t.Optional[str] = None

    # pylint: disable-next=unused-private-member
    __convert_paths = p.validator('collection_cache',
                                  pre=True, allow_reuse=True)(convert_path)


class LibContext(BaseModel):
    """
    Structure and defaults of the lib_ctx.

    :ivar chunksize: number of bytes to read or write at one time for network or file IO
    :ivar process_max: Maximum number of worker processes for parallel operations.  It may be None
        to mean, use all available CPU cores.
    :ivar thread_max: Maximum number of helper threads for parallel operations
    :ivar file_check_content: Maximum number of bytes of a file to read before writing it to
        compare contents. If contents are as expected, file is not overwritten. Set to 0 to
        disable.
    :ivar max_retries: Maximum number of times to retry an http request (in case of timeouts and
        other transient problems.
    :ivar doc_parsing_backend: The backend to use for parsing the documentation strings from
        plugins.  This is DEPRECATED and will be removed from the library context in
        antsibull-core 3.0.0.
        'auto' selects a backend depending on the ansible-core version.
        'ansible-internal' is the fastest, but does not work with ansible-core 2.13+.
        'ansible-core-2.13' is also very fast, but requires ansible-core 2.13+.
        'ansible-doc' exists in case of problems with the ansible-internal backend.
    """

    chunksize: int = 4096
    # DEPRECATED: doc_parsing_backend will be removed in antsibull-core 3.0.0
    doc_parsing_backend: str = 'auto'
    max_retries: int = 10
    process_max: t.Optional[int] = None
    thread_max: int = 8
    file_check_content: int = 262144

    # pylint: disable-next=unused-private-member
    __convert_nones = p.validator('process_max', pre=True, allow_reuse=True)(convert_none)
